/*
 *  Main authors:
 *     Fabio Rossi <rossi@dmi.unipg.it>
 *
 *  Contributing authors:
 *     Stefano Bistarelli <bista@dmi.unipg.it>
 *     Francesco Santini <francesco.santini@dmi.unipg.it>
 *
 *  Copyright:
 *     Fabio Rossi, 2015
 *     Stefano Bistarelli, 2015
 *     Francesco Santini, 2015
 *
 * Last modified: Sep 20th, 2017
 *
 *  This file is part of ConArg Project, a software for Argumentation Framework:
 *     http://www.dmi.unipg.it/conarg
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#ifndef AFENGINE_H_
#define AFENGINE_H_

#include <string>
#include <vector>
#include <set>
#include "tSolution.h"

#define CL_SRTYPE_DUNG		0
#define CL_SRTYPE_FUZZY		1
#define CL_SRTYPE_WEIGHTED	2
#define CL_SRTYPE_INVALID	3

// Gecode heuristics codes for VAR branching
#define CL_UNSPECIFIED_VAR_HEUR			INT_MIN
#define CL_WRONG_VARHEUR				INT_MIN+1
#define CL_BOOL_VAR_AFC_MAX_HEUR		1
#define CL_BOOL_VAR_CHB_MAX_HEUR		2
#define CL_BOOL_VAR_DEGREE_MAX_HEUR		3
#define CL_BOOL_VAR_RND_HEUR			4
#define CL_BOOL_VAR_AFC_MIN_HEUR		5
#define CL_BOOL_VAR_CHB_MIN_HEUR		6
#define CL_BOOL_VAR_DEGREE_MIN_HEUR		7
#define CL_BOOL_VAR_ACTION_MAX_HEUR		8
#define CL_BOOL_VAR_ACTION_MIN_HEUR		9
#define CL_BOOL_VAR_MERIT_MAX_HEUR		10
#define CL_BOOL_VAR_MERIT_MIN_HEUR		11
#define CL_BOOL_VAR_NONE_HEUR			12
#define CL_BOOL_VAR_DEFAULT_HEUR		CL_BOOL_VAR_AFC_MAX_HEUR

// Gecode heuristics codes for VAL branching
#define CL_UNSPECIFIED_VAL_HEUR			INT_MIN
#define CL_WRONG_VALHEUR				INT_MIN+1
#define CL_BOOL_VAL_MAX_HEUR			1
#define CL_BOOL_VAL_MIN_HEUR			2
#define CL_BOOL_VAL_RND_HEUR			3
#define CL_BOOL_VAL_DEFAULT_HEUR		CL_UNSPECIFIED_VAL_HEUR

// Defines for heuristics parameter
#define CL_INVALID_PARAM_HEUR			DBL_MIN
#define CL_AFC_DEFAULT_DECAY			0.99
#define CL_ACTION_DEFAULT_DECAY			0.99

using namespace std;

namespace AF {

class AFEngine {
private:
	string 			m_sInputFile;
	void			*m_pOptions;
	void 			*m_pDLFile;
	void 			*m_AFSE;
	void			FreeDLFile();
	void			FreeAFSearchEngine();
	int				m_iLastCallType;
	string GetArgName(int iIndex);
	set<tSolution *> *GetExtensionsEnum(int iExtCode);
	set<tSolution *> *GetAlphaExtensionsEnum(int iSrt, double dAlpha, double dGamma, int iExtCode);
	set<tSolution *> *GetExtensionsExist(int iExtCode);
	set<tSolution *> *GetAlphaExtensionsExist(int iSrt, double dAlpha, double dGamma, int iExtCode);
	string			 GetAcceptance(int iExtCode, int iModusOperandi, string &sArgCheck);
	string			 GetAcceptance(int iSrt, double dAlpha, double dGamma, int iExtCode, int iModusOperandi, string &sArgCheck);
	bool IsExtension(vector<string> &vecExtToCheck, int iExtCode);
	bool IsExtension(vector<string> &vecExtToCheck, int iSrt, double dAlpha, double dGamma, int iExtCode);
	bool IsExtensionBABValuesMax(vector<string> &vecExtToCheck, int iExtCode);
	bool IsExtensionBABValuesMax(vector<string> &vecExtToCheck, int iSrt, double dAlpha, double dGamma, int iExtCode);
	bool TrivialIsExtension(vector<string> &vecArgs, int iExtCode);
	bool TrivialIsExtension(int iSrt, double dAlpha, double dGamma, vector<string> &vecArgs, int iExtCode);
	set<tSolution *> *NonEmptinessExtension(int iExtCode);
	set<tSolution *> *NonEmptinessAlphaExtension(int iSrt, double dAlpha, double dGamma, int iExtCode);
	set<tSolution *> *TrivialNonEmptinessExtension(int iExtCode);
	set<tSolution *> *TrivialNonEmptinessExtension(int iSrt, double dAlpha, double dGamma, int iExtCode);

public:
	AFEngine();
	virtual ~AFEngine();

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/*
	 * Lybrary call to laod an input file. It initializes the Library search engine too.
	 */
	bool LoadFile(string &sInputFileName, int iSRType = CL_SRTYPE_DUNG);
	/*
	 * To print the extensions found by the search
	 */
	void PrintSol(set<tSolution *> *psetSol);
	/*
	 * To free the memory allocated to store the extensions found
	 */
	void FreeSol(set<tSolution *> *psetSol);
	/*
	 *To test whether the search engine is correctly initialized
	 */
	bool EngineIsInitialized();
	/*
	 * To get the extensions found count by the last search call
	 */
	int GetSolutionsCount(set<tSolution *> *psetSol);

	double	GetFuzzyBestValue();     // To get the Fuzzy semiring best value (1)
	double	GetFuzzyWorstValue();    // To get the Fuzzy semiring worst value (0)
	double	GetWeightedBestValue();  // To get the Weighted semiring best value
	double	GetWeightedWorstValue(); // To get the Weighted semiring worst value

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/*
	 * To get/set options
	 */
	void 	SetShowSolutions(bool bFlag);     // to activate/deactivate showing of extensions found. Default false
	bool 	GetShowSolutions();
	void	SetRequestedSolutions(long lRS);  // to request n extension. Default ALL
	long	GetRequestedSolutions();
	void	SetRequesteAllSolutions();        // to request all extensions. Default true (ALL)
	void	SetStatistics(bool bStat);        // to request the CP search statistics printing. Statistics are available
											  // for DUNG and for the following semantics only: conflictfree, admissible,
	         	 	 	 	 	 	 	 	  // complete, grounded, preferred, stable, semistable, stage.
											  // For other semiring/semantics the result is unpredictable. Default false
	bool	GetStatistics();
	void	SetProboOutput(bool bProbo);      // to activate/deactivate the PROBO output format. Default true
	bool	GetProboOutput();
	void	SetPercentOutput(bool bPercent);  // to activate/deactivate the printing of percentages of arguments in all extensions.
											  // Default false.
	bool	GetPercentOutput();
	void	SetSilentMode(bool bSilentMode);  // to set/get silent (not verbose) mode. Default true
	bool	GetSilentMode();
	void	SetShowSolutionsCount(bool bShowSolutionsCount);   // to activate/deactivate showing of extensions found count
	bool	GetShowSolutionsCount();
	void	SetSemiringType(int iSRType);     // to set/get the current semiring. Default DUNG
	int		GetSemiringType();
	void	SetVarHeur(int iVarHeur);		  // to set Gecode heuristic for Variable branching
	void	SetDefaultVarHeur();
	int		GetVarHeur();
	void	SetValHeur(int iValHeur);		  // to set Gecode heuristic for Variable Value branching
	void	SetDefaultValHeur();
	int		GetValHeur();
	void	SetHeurParam(double dParam);	  // to set the decay factor for AFC and Action Gecode heuristics.
											  // Ignored for other heuristics
	double	GetHeurParam();
	void	SetThreadsCount(int iThreadsCount); // to set the number of requested threads for Gecode search. Default 1
	int		GetThreadsCount();
	void	ResetDefaults();    // resets to default options values

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/*
	 * To enumerate extensions, test for existence, acceptance, verification and non-emptyness
	 */
	// ONLY DUNG (crisp semantics)
	set<tSolution *> *GetAdmissibleEnum();
	set<tSolution *> *GetAdmissibleExist();
	string GetAdmissibleCred(string &sArg);
	string GetAdmissibleScept(string &sArg);
	bool   IsAdmissible(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessAdmissible();

	set<tSolution *> *GetStableEnum();
	set<tSolution *> *GetStableExist();
	string GetStableCred(string &sArg);
	string GetStableScept(string &sArg);
	bool   IsStable(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessStable();

	set<tSolution *> *GetConflictFreeEnum();
	set<tSolution *> *GetConflictFreeExist();
	string GetConflictFreeCred(string &sArg);
	string GetConflictFreeScept(string &sArg);
	bool   IsConflictFree(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessConflictFree();

	set<tSolution *> *GetCompleteEnum();
	set<tSolution *> *GetCompleteExist();
	string GetCompleteCred(string &sArg);
	string GetCompleteScept(string &sArg);
	bool   IsComplete(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessComplete();

	set<tSolution *> *GetGroundedEnum();
	set<tSolution *> *GetGroundedExist();
	string GetGroundedCred(string &sArg);
	string GetGroundedScept(string &sArg);
	bool   IsGrounded(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessGrounded();

	set<tSolution *> *GetPreferredEnum();
	set<tSolution *> *GetPreferredExist();
	string GetPreferredCred(string &sArg);
	string GetPreferredScept(string &sArg);
	bool   IsPreferred(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessPreferred();

	set<tSolution *> *GetIdealEnum();
	set<tSolution *> *GetIdealExist();
	string GetIdealCred(string &sArg);
	string GetIdealScept(string &sArg);
	bool   IsIdeal(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessIdeal();

	set<tSolution *> *GetSemiStableEnum();
	set<tSolution *> *GetSemiStableExist();
	string GetSemiStableCred(string &sArg);
	string GetSemiStableScept(string &sArg);
	bool   IsSemiStable(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessSemiStable();

	set<tSolution *> *GetStageEnum();
	set<tSolution *> *GetStageExist();
	string GetStageCred(string &sArg);
	string GetStageScept(string &sArg);
	bool   IsStage(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessStage();

	set<tSolution *> *GetEagerEnum();
	set<tSolution *> *GetEagerExist();
	string GetEagerCred(string &sArg);
	string GetEagerScept(string &sArg);
	bool   IsEager(vector<string> &vecExtToCheck);
	set<tSolution *> *NonEmptinessEager();

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// ALPHA Semantics (other semirings than DUNG)
	set<tSolution *> *GetAdmissibleEnum(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetAdmissibleEnum(double dAlpha, double dGamma);
	set<tSolution *> *GetAdmissibleExist(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetAdmissibleExist(double dAlpha, double dGamma);
	string GetAdmissibleCred(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetAdmissibleCred(string &sArg, double dAlpha, double dGamma);
	string GetAdmissibleScept(string &sArg, double dAlpha, double dGamma);
	string GetAdmissibleScept(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	bool   IsAdmissible(vector<string> &vecExtToCheck, double dAlpha, double dGamma);
	bool   IsAdmissible(vector<string> &vecExtToCheck, int semiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessAdmissible(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessAdmissible(double dAlpha, double dGamma);

	set<tSolution *> *GetConflictFreeEnum(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetConflictFreeEnum(double dAlpha, double dGamma);
	set<tSolution *> *GetConflictFreeExist(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetConflictFreeExist(double dAlpha, double dGamma);
	string GetConflictFreeCred(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetConflictFreeCred(string &sArg, double dAlpha, double dGamma);
	string GetConflictFreeScept(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetConflictFreeScept(string &sArg, double dAlpha, double dGamma);
	bool   IsConflictFree(vector<string> &vecExtToCheck, double dAlpha, double dGamma);
	bool   IsConflictFree(vector<string> &vecExtToCheck, int semiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessConflictFree(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessConflictFree(double dAlpha, double dGamma);

	set<tSolution *> *GetCompleteEnum(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetCompleteEnum(double dAlpha, double dGamma);
	set<tSolution *> *GetCompleteExist(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetCompleteExist(double dAlpha, double dGamma);
	string GetCompleteCred(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetCompleteScept(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetCompleteCred(string &sArg, double dAlpha, double dGamma);
	string GetCompleteScept(string &sArg, double dAlpha, double dGamma);
	bool   IsComplete(vector<string> &vecExtToCheck, double dAlpha, double dGamma);
	bool   IsComplete(vector<string> &vecExtToCheck, int semiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessComplete(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessComplete(double dAlpha, double dGamma);

	set<tSolution *> *GetGroundedEnum(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetGroundedEnum(double dAlpha, double dGamma);
	set<tSolution *> *GetGroundedExist(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetGroundedExist(double dAlpha, double dGamma);
	string GetGroundedCred(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetGroundedCred(string &sArg, double dAlpha, double dGamma);
	string GetGroundedScept(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetGroundedScept(string &sArg, double dAlpha, double dGamma);
	bool   IsGrounded(vector<string> &vecExtToCheck, double dAlpha, double dGamma);
	bool   IsGrounded(vector<string> &vecExtToCheck, int semiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessGrounded(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessGrounded(double dAlpha, double dGamma);

	set<tSolution *> *GetPreferredEnum(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetPreferredEnum(double dAlpha, double dGamma);
	set<tSolution *> *GetPreferredExist(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetPreferredExist(double dAlpha, double dGamma);
	string GetPreferredCred(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetPreferredCred(string &sArg, double dAlpha, double dGamma);
	string GetPreferredScept(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetPreferredScept(string &sArg, double dAlpha, double dGamma);
	bool IsPreferred(vector<string> &ExtToCheck, int iSemiringType, double dAlpha, double dGamma);
	bool IsPreferred(vector<string> &ExtToCheck, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessPreferred(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessPreferred(double dAlpha, double dGamma);

	set<tSolution *> *GetStableEnum(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetStableEnum(double dAlpha, double dGamma);
	set<tSolution *> *GetStableExist(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *GetStableExist(double dAlpha, double dGamma);
	string GetStableCred(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetStableCred(string &sArg, double dAlpha, double dGamma);
	string GetStableScept(string &sArg, int iSemiringType, double dAlpha, double dGamma);
	string GetStableScept(string &sArg, double dAlpha, double dGamma);
	bool IsStable(vector<string> &ExtToCheck, int iSemiringType, double dAlpha, double dGamma);
	bool IsStable(vector<string> &ExtToCheck, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessStable(int iSemiringType, double dAlpha, double dGamma);
	set<tSolution *> *NonEmptinessStable(double dAlpha, double dGamma);
};

} /* namespace AF */

#endif /* AFENGINE_H_ */
