/*
 *  Main authors:
 *     Fabio Rossi <rossi@dmi.unipg.it>
 *
 *  Contributing authors:
 *     Stefano Bistarelli <bista@dmi.unipg.it>
 *     Francesco Santini <francesco.santini@dmi.unipg.it>
 *
 *  Copyright:
 *     Fabio Rossi, 2015
 *     Stefano Bistarelli, 2015
 *     Francesco Santini, 2015
 *
 * Last modified: Oct 24th, 2017
 *
 *  This file is part of ConArg Project, a software for Argumentation Framework:
 *     http://www.dmi.unipg.it/conarg
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <iostream>
#include <string>
#include <vector>
#include <stdio.h>
#include <stdlib.h>

/*****************************************************************************************************************/
/*
 * Including the Library's header file
 */
#include "AFEngine.h"

using namespace std;
using namespace AF;

// main function
int main(int argc, char* argv[]) {
	int iRet = 0;

	try {
		/*****************************************************************************************************************/
		/*
		 * Creating an objet of class AFEngine.
		 * This is the Library object and it is the only object to be created
		 */
		AFEngine *af = new AFEngine();

		/*
		 * Declaring a STL set to store the extensions found.
		 *
		 * The tSolution struct is defined in tSolution.h:
			struct tSolution {
				long lArgsSize;         the number of arguments in the AF problem
				bool *pArgs;            the values of the arguments (true or false) in the extension
				double dCost;			the cost of the extensnsion (ignored for DUNG semiring)

				tSolution(long lArgsCount)    the constructor

				~tSolution()     the destructor

				tSolution &operator=(const tSolution &other)      the '=' operator

				long GetSolCard()      returns the cardinality of the extension, i.e. the count of arguments true
			};
		 */
		set<tSolution *> *tsetExtensions;

		/*
		 * Declaring a STL vector to store a set of arguments to be verified whether it is an extension
		 */
		vector<string> vecArgs;

		/*
		 * Loading test.dl file that contains an argumentation problem definition:
		 	% arguments
			arg(a).
			arg(b).
			arg(c).
			arg(d).
			arg(e).

			% attack relations
			att(a,b).
			att(c,b).
			att(c,d).
			att(d,c).
			att(d,e).
			att(e,e).

		  *	LoadFile returns true if there are no errors, false otherwise. It initializes the Library search engine too.
		  *	To test if the searcg engine is initialized use:
		  *	bool EngineIsInitialized()
		 */
		string sFileName = "test.dl";
		cout << "Loading \"test.dl\" file (DUNG semiring)...." << endl;
		if(af->LoadFile(sFileName)) {

			/*****************************************************************************************************************/
			/*
			 * Getting all (DUNG) Admissible extensions.
			 * The extensions found are stored in the variable tsetExtensions.
			 *
			 * In general the functions to enumerate extensions are in the form (for Admissible semantic):
			 *
			 * set<tSolution *> *GetAdmissibleEnum()
			 * 	it considers DUNG semiring only
			 *
			 * set<tSolution *> *GetAdmissibleEnum(int iSRType, double dAlpha, double dGamma)
			 * 	for all semirings. If the semiring specified is DUNG dAlpha and dGamma are ignored.
			 *
			 * set<tSolution *> *GetAdmissibleEnum(double dAlpha, double dGamma)
			 * 	for the current semiring set. If the current semiring is DUNG dAlpha and dGamma are ignored.
			 *
			 * 'iSRType' can be one of the following defined (in AFEngine.h) values;
			 * #define CL_SRTYPE_DUNG		0
			 * #define CL_SRTYPE_FUZZY		1
			 * #define CL_SRTYPE_WEIGHTED	2
			 *			 *
			 * Analogous functions exist for other semantics. E.g. for Complete:
			 * set<tSolution *> *GetCompleteEnum()
			 * set<tSolution *> *GetCompleteEnum(int iSRType, double dAlpha, double dGamma)
			 * set<tSolution *> *GetCompleteEnum(double dAlpha, double dGamma)

			 * For a complete list of enumeration (and ALL other library) functions see the 'public' section of AFEngine.h
			 *
			 * The default semiring is CL_SRTYPE_DUNG and the current one can be set by
			 * void	SetSemiringType(int iSRType)
			 *
			 * The current semiring can be determined by calling:
			 * int GetSemiringType()
			 */
			cout << "Getting all (DUNG) admissible extensions...." << endl;
			/*
			 * Default of the option 'showing extensions found' is false,
			 * indeed no extensions are printed by GetAdmissibleEnum().
			 */
			tsetExtensions = af->GetAdmissibleEnum();

			/*
			 * Printing the extensions found.
			 * The default output format is PROBO (see ICCMA15 competition).
			 *
			 * To activate/deactivate/check PROBO option use:
			 * void	SetProboOutput(bool bProbo);
			 * bool	GetProboOutput();
			 */
			af->PrintSol(tsetExtensions);


			/*
			 * Counting and printing the number of solutions found
			 */
			cout << af->GetSolutionsCount(tsetExtensions) << " solution(s) found" << endl;

			/*
			 * Freeing the memory allocated to store the extensions found
			 */
			af->FreeSol(tsetExtensions);


			/*****************************************************************************************************************/
			/*
			 * Getting all Complete extensions. Showing extensions found is activated before to compute the extensions so
			 * We do not need to call PrintSol(tsetExtensions).
			 * However the extensions are stored in the variable tsetExtensions.
			 *
			 * Functions to get/set current value of 'Showing extensions found':
			 * void 	SetShowSolutions(bool bFlag)
			 * bool 	GetShowSolutions()
			 */
			cout << endl << "Getting all complete extensions...." << endl;
			af->SetShowSolutions(true);
			tsetExtensions = af->GetCompleteEnum();

			/*
			 * Freeing anyway the memory allocated to store solutions
			 */
			af->FreeSol(tsetExtensions);
			af->SetShowSolutions(false);  // Resetting 'Showing extensions found' to false (default value)


			/*****************************************************************************************************************/
			/*
			 * Getting all Preferred extensions.
			 * Using CHB_MAX heuristic is requested before starting search. See Gecode manual (www.gecode.org) for details on
			 * heuristics.
			 *
			 * The current heuristic for branching on variables can be specified by calling:
			 * void SetVarHeur(int iVarHeuristicCode)
			 *
			 * where iVarHeuristicCode can be:
			 * #define CL_BOOL_VAR_AFC_MAX_HEUR			1
			 * #define CL_BOOL_VAR_CHB_MAX_HEUR			2
			 * #define CL_BOOL_VAR_DEGREE_MAX_HEUR		3
			 * #define CL_BOOL_VAR_RND_HEUR				4
			 * #define CL_BOOL_VAR_AFC_MIN_HEUR			5
			 * #define CL_BOOL_VAR_CHB_MIN_HEUR			6
			 * #define CL_BOOL_VAR_DEGREE_MIN_HEUR		7
			 * #define CL_BOOL_VAR_ACTION_MAX_HEUR		8
			 * #define CL_BOOL_VAR_ACTION_MIN_HEUR		9
			 * #define CL_BOOL_VAR_MERIT_MAX_HEUR		10
			 * #define CL_BOOL_VAR_MERIT_MIN_HEUR		11
			 * #define CL_BOOL_VAR_DEFAULT_HEUR			CL_BOOL_VAR_AFC_MAX_HEUR
			 *
			 * To set default current variable branching heuristic call:
			 * void	SetDefaultVarHeur();
			 *
			 * The current heuristic for branching on variables value can be specified by calling:
			 * void SetValHeur(int iVaHeuristicCode)
			 *
			 * where iValHeuristicCode can be:
			 * #define CL_UNSPECIFIED_VAL_HEUR		INT_MIN
			 * #define CL_BOOL_VAL_MAX_HEUR			1
			 * #define CL_BOOL_VAL_MIN_HEUR			2
			 * #define CL_BOOL_VAL_RND_HEUR			3
			 * #define CL_BOOL_VAL_DEFAULT_HEUR		CL_UNSPECIFIED_VAL_HEUR
			 *
			 * To set default current variables value branching heuristic call:
			 * void	SetDefaultValHeur();
			 *
			 * Action and AFC heuristics accept a decay factor d (0 < d <= 1) as parameter for tuning.
			 * Default values for d are:
			 * #define CL_AFC_DEFAULT_DECAY				0.99
			 * #define CL_ACTION_DEFAULT_DECAY			0.99
			 * and You can set d by calling:
			 *
			 * void	SetHeurParam(double dParam);
			 *
			 */
			cout << endl << "Getting all preferred extensions using CHB_MAX ...." << endl;
			af->SetVarHeur(CL_BOOL_VAR_CHB_MAX_HEUR);
			tsetExtensions = af->GetPreferredEnum();
			af->PrintSol(tsetExtensions);
			cout << af->GetSolutionsCount(tsetExtensions) << " solution(s) found" << endl;
			af->FreeSol(tsetExtensions);
			af->SetDefaultVarHeur();  // To reset variables branching heuristic to the default value


			/*****************************************************************************************************************/
			/*
			 * Checking the Credulous Acceptance of the argument "b" for the stable semantic.
			 *
			 * Functions to check acceptance are in the form:
			 * string GetStableCred(string &sArgToCheck)       											CL_SRTYPE_DUNG only
			 * string GetSemiStableScept(string &sArgToCheck)											CL_SRTYPE_DUNG only
			 * string GetPreferredCred(string &sArg, int iSemiringType, double dAlpha, double dGamma)	for other semirings
			 * string GetStableCred(string &sArg, double dAlpha, double dGamma)							for the current semiring
			 * string GetStableScept(string &sArg, int iSemiringType, double dAlpha, double dGamma)		for other semirings
			 * string GetConflictFreeScept(string &sArg, double dAlpha, double dGamma)					for the current semiring
			 *
			 * These functions return "YES" or "NO".
			 *
			 * For a complete list of acceptance (and ALL other library) functions see the 'public' section of AFEngine.h
			 */
			cout << endl << "Checking the Credulous Acceptance of the argument \"b\" for the stable semantic...." << endl;
			string s = "b";
			cout << af->GetStableCred(s) << endl;


			/*****************************************************************************************************************/
			/*
			 * Checking the existence of a preferred extension.
			 *
			 * GetPreferredExist() returns a pointer to a set<tSolution *>. The set contains the extension found or is empty
			 * if no solutions exist.
			 *
			 * Functions to check existence are in the form:
			 * set<tSolution *> *GetIdealExist();													CL_SRTYPE_DUNG only
			 * set<tSolution *> *GetStableExist(int iSemiringType, double dAlpha, double dGamma)	for other semirings
			 * set<tSolution *> *GetStableExist(double dAlpha, double dGamma)						for the current semiring
			 *
			 * For a complete list of existence (and ALL other library) functions see the 'public' section of AFEngine.h
			 */
			cout << endl << "Checking the existence of preferred extensions...." << endl;
			tsetExtensions = af->GetPreferredExist();
			af->PrintSol(tsetExtensions);
			af->FreeSol(tsetExtensions);


			/*****************************************************************************************************************/
			/*
			 * Verifying if {b,c} is a complete extension.
			 *
			 * First a STL vector representing the extension to check is built, then IsComplete(vecArgs) is called.
			 *
			 * Functions to verify extension are in the form:
			 * bool   IsEager(vector<string> &vecExtToCheck)													CL_SRTYPE_DUNG only
			 * bool   IsGrounded(vector<string> &vecExtToCheck, double dAlpha, double dGamma)					for the current semiring
			 * bool   IsGrounded(vector<string> &vecExtToCheck, int semiringType, double dAlpha, double dGamma)	for other semirings than DUNG
			 *
			 * For a complete list of verifying extension (and ALL other library) functions see the 'public' section of AFEngine.h
			 */
			cout << endl << "Verifying if {b,c} is a complete extension...." << endl;
			/*
			 * Filling the vector that contains the set of arguments to check
			 */
			vecArgs.clear();
			vecArgs.push_back("b");
			vecArgs.push_back("c");

			/*
			 * Checking if {b,c} is a complete extension
			 */
			if(af->IsComplete(vecArgs))
				cout << "{b,c} is a complete extension" << endl;
			else
				cout << "{b,c} is not a complete extension" << endl;


			/*****************************************************************************************************************/
			/*
			 * Verifying admissible non-emptiness
			 *
			 * NonEmptinessAdmissible() returns a pointer to a set<tSolution *>. The set contains the non empty extension
			 * found or is empty if non-empty extensions do not exist.
			 *
			 * Non-emptyness functions are in the form:
			 * set<tSolution *> *NonEmptinessIdeal()													CL_SRTYPE_DUNG only
			 * set<tSolution *> *NonEmptinessGrounded(int iSemiringType, double dAlpha, double dGamma)	for other semirings
			 * set<tSolution *> *NonEmptinessGrounded(double dAlpha, double dGamma)						for the current semiring
			 *
			 * For a complete list of non-emptyness extension (and ALL other library) functions see the 'public' section of AFEngine.h
			 */
			cout << endl << "Verifying admissible non-emptiness...." << endl;
			tsetExtensions = af->NonEmptinessAdmissible();
			af->PrintSol(tsetExtensions);
			af->FreeSol(tsetExtensions);


			/*****************************************************************************************************************/
			/*
			 * Loading a Weighted semiring input file.
			 */
			sFileName = "erdosrenyiweighted11.dl";
			cout << endl << "Loading \"" << sFileName << "\" file (Weighted semiring)...." << endl;
			if(af->LoadFile(sFileName)) {

				/*****************************************************************************************************************/
				/*
				 * Getting all Admissible extension for the weighted semiring, alpha=3 and gamma=0.
				 * First the current semiring is set to CL_SRTYPE_WEIGHTED
				 */
				cout << endl << "Getting all Weighted alpha=3, gamma=0 admissible extensions...." << endl;

				/*
				 * Setting current semiring
				 */
				af->SetSemiringType(CL_SRTYPE_WEIGHTED);

				/*
				 * Getting admissible extensions
				 */
				tsetExtensions = af->GetAdmissibleEnum(3, 0);

				/*
				 * Printing the extensions found
				 */
				af->PrintSol(tsetExtensions);

				/*
				 * Counting (and printing) the solutions found
				 */
				cout << af->GetSolutionsCount(tsetExtensions) << " solution(s) found" << endl;

				/*
				 * Freeing memory allocated to store the extensions found
				 */
				af->FreeSol(tsetExtensions);

				/*****************************************************************************************************************/
				/*
				 * Verifying complete non-emptyness for alpha=3 and gamma=0. The current semiring is still CL_SRTYPE_WEIGHTED
				 */
				cout << endl << "Verifying complete non-emptiness...." << endl;
				tsetExtensions = af->NonEmptinessAdmissible(3, 0);
				af->PrintSol(tsetExtensions);
				af->FreeSol(tsetExtensions);

				/*****************************************************************************************************************/
				// Have fun!
			}
			else {
				iRet = -1;
			}
		}
		else {
			iRet = -1;
		}

		delete af;
	} catch (exception &e) {
		std::cerr << "Exception: " << e.what() << std::endl;
		iRet = -1;
	}

	return iRet;
}




